"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.runCodeInTestContext = runCodeInTestContext;
exports.default = _default;

function babel() {
  var data = _interopRequireWildcard(require("@babel/core"));

  babel = function babel() {
    return data;
  };

  return data;
}

function _helperFixtures() {
  var data = _interopRequireDefault(require("@babel/helper-fixtures"));

  _helperFixtures = function _helperFixtures() {
    return data;
  };

  return data;
}

function _sourceMap() {
  var data = _interopRequireDefault(require("source-map"));

  _sourceMap = function _sourceMap() {
    return data;
  };

  return data;
}

function _codeFrame() {
  var data = require("@babel/code-frame");

  _codeFrame = function _codeFrame() {
    return data;
  };

  return data;
}

function _defaults() {
  var data = _interopRequireDefault(require("lodash/defaults"));

  _defaults = function _defaults() {
    return data;
  };

  return data;
}

function _includes() {
  var data = _interopRequireDefault(require("lodash/includes"));

  _includes = function _includes() {
    return data;
  };

  return data;
}

var helpers = _interopRequireWildcard(require("./helpers"));

function _extend() {
  var data = _interopRequireDefault(require("lodash/extend"));

  _extend = function _extend() {
    return data;
  };

  return data;
}

function _merge() {
  var data = _interopRequireDefault(require("lodash/merge"));

  _merge = function _merge() {
    return data;
  };

  return data;
}

function _resolve() {
  var data = _interopRequireDefault(require("resolve"));

  _resolve = function _resolve() {
    return data;
  };

  return data;
}

function _assert() {
  var data = _interopRequireDefault(require("assert"));

  _assert = function _assert() {
    return data;
  };

  return data;
}

function _fs() {
  var data = _interopRequireDefault(require("fs"));

  _fs = function _fs() {
    return data;
  };

  return data;
}

function _path() {
  var data = _interopRequireDefault(require("path"));

  _path = function _path() {
    return data;
  };

  return data;
}

function _vm() {
  var data = _interopRequireDefault(require("vm"));

  _vm = function _vm() {
    return data;
  };

  return data;
}

function _jestDiff() {
  var data = _interopRequireDefault(require("jest-diff"));

  _jestDiff = function _jestDiff() {
    return data;
  };

  return data;
}

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) { var desc = Object.defineProperty && Object.getOwnPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : {}; if (desc.get || desc.set) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } } newObj.default = obj; return newObj; } }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; var ownKeys = Object.keys(source); if (typeof Object.getOwnPropertySymbols === 'function') { ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) { return Object.getOwnPropertyDescriptor(source, sym).enumerable; })); } ownKeys.forEach(function (key) { _defineProperty(target, key, source[key]); }); } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var moduleCache = {};

var testContext = _vm().default.createContext(_objectSpread({}, helpers, {
  process: process,
  transform: babel().transform,
  setTimeout: setTimeout,
  setImmediate: setImmediate,
  expect: expect
}));

testContext.global = testContext;
runModuleInTestContext("@babel/polyfill", __filename);
runCodeInTestContext((0, babel().buildExternalHelpers)());

function runModuleInTestContext(id, relativeFilename) {
  var filename = _resolve().default.sync(id, {
    basedir: _path().default.dirname(relativeFilename)
  });

  if (filename === id) return require(id);
  if (moduleCache[filename]) return moduleCache[filename].exports;
  var module = moduleCache[filename] = {
    id: filename,
    exports: {}
  };

  var dirname = _path().default.dirname(filename);

  var req = function req(id) {
    return runModuleInTestContext(id, filename);
  };

  var src = _fs().default.readFileSync(filename, "utf8");

  var code = "(function (exports, require, module, __filename, __dirname) {" + src + "\n});";

  _vm().default.runInContext(code, testContext, {
    filename: filename,
    displayErrors: true
  }).call(module.exports, module.exports, req, module, filename, dirname);

  return module.exports;
}

function runCodeInTestContext(code, opts) {
  if (opts === void 0) {
    opts = {};
  }

  var filename = opts.filename || null;
  var dirname = filename ? _path().default.dirname(filename) : null;
  var req = filename ? function (id) {
    return runModuleInTestContext(id, filename);
  } : null;
  var module = {
    id: filename,
    exports: {}
  };
  var src = "(function(exports, require, module, __filename, __dirname, opts) {" + code + "\n});";
  return _vm().default.runInContext(src, testContext, {
    filename: filename,
    displayErrors: true
  })(module.exports, req, module, filename, dirname, opts);
}

function wrapPackagesArray(type, names, optionsDir) {
  return (names || []).map(function (val) {
    if (typeof val === "string") val = [val];

    if (val[0][0] === ".") {
      if (!optionsDir) {
        throw new Error("Please provide an options.json in test dir when using a " + "relative plugin path.");
      }

      val[0] = _path().default.resolve(optionsDir, val[0]);
    } else {
      val[0] = __dirname + "/../../babel-" + type + "-" + val[0];
    }

    return val;
  });
}

function checkDuplicatedNodes(ast) {
  var nodes = new WeakSet();
  var parents = new WeakMap();

  var setParent = function setParent(child, parent) {
    if (typeof child === "object" && child !== null) {
      var p = parents.get(child);

      if (!p) {
        p = [];
        parents.set(child, p);
      }

      p.unshift(parent);
    }
  };

  var registerChildren = function registerChildren(node) {
    for (var key in node) {
      if (Array.isArray(node[key])) {
        node[key].forEach(function (child) {
          return setParent(child, node);
        });
      } else {
        setParent(node[key], node);
      }
    }
  };

  var parentIs = function parentIs(node, test) {
    return (parents.get(node) || []).some(function (parent) {
      return test(parent);
    });
  };

  var isByRegenerator = function isByRegenerator(node) {
    if (!node) {
      return false;
    } else if (node.type === "Identifier") {
      if (/^_(?:context|value|callee|marked)\d*$/.test(node.name)) {
        return true;
      } else if (/^t\d+$/.test(node.name) && parentIs(node, function (parent) {
        return parent.type === "MemberExpression" && isByRegenerator(parent.object);
      })) {
        return true;
      } else if (parentIs(node, function (parent) {
        return parent.type === "VariableDeclarator" && parentIs(parent, function (parent) {
          return parent.type === "VariableDeclaration" && parentIs(parent, function (parent) {
            return parent.type === "BlockStatement" && parentIs(parent, function (parent) {
              return parent.type === "FunctionExpression" && isByRegenerator(parent.id);
            });
          });
        });
      })) {
        return true;
      } else if (parentIs(node, function (parent) {
        return parent.type === "VariableDeclarator" && parentIs(parent, function (parent) {
          return parent.type === "VariableDeclaration" && parentIs(parent, function (parent) {
            return parent.type === "BlockStatement" && parent.body.length === 2 && parent.body[1].type === "ReturnStatement" && parent.body[1].argument.type === "CallExpression" && parent.body[1].argument.callee.type === "MemberExpression" && parent.body[1].argument.callee.property.type === "Identifier" && parent.body[1].argument.callee.property.name === "wrap";
          });
        });
      })) {
        return true;
      } else if (parentIs(node, function (parent) {
        return parent.type === "CallExpression" && parent.arguments.length === 3 && parent.arguments[1] === node && parent.callee.type === "MemberExpression" && parent.callee.property.type === "Identifier" && parent.callee.property.name === "wrap";
      })) {
        return true;
      } else if (parentIs(node, function (parent) {
        return parent.type === "CallExpression" && parent.callee.type === "MemberExpression" && parent.callee.property.type === "Identifier" && parent.callee.property.name === "mark";
      })) {
        return true;
      }
    } else if (node.type === "MemberExpression") {
      return isByRegenerator(node.object);
    } else if (node.type === "CallExpression") {
      return isByRegenerator(node.callee);
    } else if (node.type === "AssignmentExpression") {
      return isByRegenerator(node.left);
    } else if (node.type === "NumericLiteral") {
      if (parentIs(node, function (parent) {
        return parent.type === "AssignmentExpression" && isByRegenerator(parent.left);
      })) {
        return true;
      } else if (parentIs(node, function (parent) {
        return parent.type === "CallExpression" && parent.callee.type === "MemberExpression" && isByRegenerator(parent.callee.object);
      })) {
        return true;
      }
    }

    return false;
  };

  var hidePrivateProperties = function hidePrivateProperties(key, val) {
    if (key[0] === "_") return "[Private]";
    return val;
  };

  babel().types.traverseFast(ast, function (node) {
    registerChildren(node);
    if (isByRegenerator(node)) return;

    if (nodes.has(node)) {
      throw new Error("Do not reuse nodes. Use `t.cloneNode` to copy them.\n" + JSON.stringify(node, hidePrivateProperties, 2) + "\nParent:\n" + JSON.stringify(parents.get(node), hidePrivateProperties, 2));
    }

    nodes.add(node);
  });
}

function run(task) {
  var actual = task.actual;
  var expected = task.expect;
  var exec = task.exec;
  var opts = task.options;
  var optionsDir = task.optionsDir;

  function getOpts(self) {
    var newOpts = (0, _merge().default)({
      filename: self.loc,
      filenameRelative: self.filename,
      sourceFileName: self.filename,
      sourceType: "script",
      babelrc: false
    }, opts);
    newOpts.plugins = wrapPackagesArray("plugin", newOpts.plugins, optionsDir);
    newOpts.presets = wrapPackagesArray("preset", newOpts.presets, optionsDir).map(function (val) {
      if (val.length > 2) {
        throw new Error("Unexpected extra options " + JSON.stringify(val.slice(2)) + " passed to preset.");
      }

      return val;
    });
    return newOpts;
  }

  var execCode = exec.code;
  var result;
  var resultExec;

  if (execCode) {
    var execOpts = getOpts(exec);
    result = babel().transform(execCode, execOpts);
    checkDuplicatedNodes(result.ast);
    execCode = result.code;

    try {
      resultExec = runCodeInTestContext(execCode, execOpts);
    } catch (err) {
      err.message = exec.loc + ": " + err.message + "\n" + (0, _codeFrame().codeFrameColumns)(execCode, {});
      throw err;
    }
  }

  var actualCode = actual.code;
  var expectCode = expected.code;

  if (!execCode || actualCode) {
    result = babel().transform(actualCode, getOpts(actual));
    checkDuplicatedNodes(result.ast);

    if (!expected.code && result.code && !opts.throws && _fs().default.statSync(_path().default.dirname(expected.loc)).isDirectory() && !process.env.CI) {
      var expectedFile = expected.loc.replace(/\.m?js$/, result.sourceType === "module" ? ".mjs" : ".js");
      console.log("New test file created: " + expectedFile);

      _fs().default.writeFileSync(expectedFile, result.code + "\n");

      if (expected.loc !== expectedFile) {
        try {
          _fs().default.unlinkSync(expected.loc);
        } catch (e) {}
      }
    } else {
      actualCode = result.code.trim();

      try {
        expect(actualCode).toEqualFile({
          filename: expected.loc,
          code: expectCode
        });
      } catch (e) {
        if (!process.env.OVERWRITE) throw e;
        console.log("Updated test file: " + expected.loc);

        _fs().default.writeFileSync(expected.loc, result.code + "\n");
      }

      if (actualCode) {
        expect(expected.loc).toMatch(result.sourceType === "module" ? /\.mjs$/ : /\.js$/);
      }
    }
  }

  if (task.sourceMap) {
    expect(result.map).toEqual(task.sourceMap);
  }

  if (task.sourceMappings) {
    var consumer = new (_sourceMap().default.SourceMapConsumer)(result.map);
    task.sourceMappings.forEach(function (mapping) {
      var actual = mapping.original;
      var expected = consumer.originalPositionFor(mapping.generated);
      expect({
        line: expected.line,
        column: expected.column
      }).toEqual(actual);
    });
  }

  if (execCode && resultExec) {
    return resultExec;
  }
}

var toEqualFile = function toEqualFile() {
  return {
    compare: function compare(actual, _ref) {
      var filename = _ref.filename,
          code = _ref.code;
      var pass = actual === code;
      return {
        pass: pass,
        message: function message() {
          var diffString = (0, _jestDiff().default)(code, actual, {
            expand: false
          });
          return "Expected " + filename + " to match transform output.\n" + "To autogenerate a passing version of this file, delete the file and re-run the tests.\n\n" + ("Diff:\n\n" + diffString);
        }
      };
    },
    negativeCompare: function negativeCompare() {
      throw new Error("Negation unsupported");
    }
  };
};

function _default(fixturesLoc, name, suiteOpts, taskOpts, dynamicOpts) {
  if (suiteOpts === void 0) {
    suiteOpts = {};
  }

  if (taskOpts === void 0) {
    taskOpts = {};
  }

  var suites = (0, _helperFixtures().default)(fixturesLoc);

  var _loop2 = function _loop2() {
    if (_isArray) {
      if (_i >= _iterator.length) return "break";
      _ref2 = _iterator[_i++];
    } else {
      _i = _iterator.next();
      if (_i.done) return "break";
      _ref2 = _i.value;
    }

    var testSuite = _ref2;
    if ((0, _includes().default)(suiteOpts.ignoreSuites, testSuite.title)) return "continue";
    describe(name + "/" + testSuite.title, function () {
      jest.addMatchers({
        toEqualFile: toEqualFile
      });

      var _loop4 = function _loop4() {
        if (_isArray2) {
          if (_i2 >= _iterator2.length) return "break";
          _ref3 = _iterator2[_i2++];
        } else {
          _i2 = _iterator2.next();
          if (_i2.done) return "break";
          _ref3 = _i2.value;
        }

        var task = _ref3;

        if ((0, _includes().default)(suiteOpts.ignoreTasks, task.title) || (0, _includes().default)(suiteOpts.ignoreTasks, testSuite.title + "/" + task.title)) {
          return "continue";
        }

        it(task.title, !task.disabled && function () {
          function runTask() {
            run(task);
          }

          (0, _defaults().default)(task.options, {
            sourceMap: !!(task.sourceMappings || task.sourceMap)
          });
          (0, _extend().default)(task.options, taskOpts);
          if (dynamicOpts) dynamicOpts(task.options, task);
          var throwMsg = task.options.throws;

          if (throwMsg) {
            delete task.options.throws;

            _assert().default.throws(runTask, function (err) {
              return throwMsg === true || err.message.indexOf(throwMsg) >= 0;
            });
          } else {
            if (task.exec.code) {
              var result = run(task);

              if (result && typeof result.then === "function") {
                return result;
              }
            } else {
              runTask();
            }
          }
        });
      };

      _loop3: for (var _iterator2 = testSuite.tests, _isArray2 = Array.isArray(_iterator2), _i2 = 0, _iterator2 = _isArray2 ? _iterator2 : _iterator2[Symbol.iterator]();;) {
        var _ref3;

        var _ret2 = _loop4();

        switch (_ret2) {
          case "break":
            break _loop3;

          case "continue":
            continue;
        }
      }
    });
  };

  _loop: for (var _iterator = suites, _isArray = Array.isArray(_iterator), _i = 0, _iterator = _isArray ? _iterator : _iterator[Symbol.iterator]();;) {
    var _ref2;

    var _ret = _loop2();

    switch (_ret) {
      case "break":
        break _loop;

      case "continue":
        continue;
    }
  }
}